import java.util.*;


/*****************************************************************************
 * This class runs expectation-maximization for the problem posed by the article
 * from (Do and Batzoglou, 2008).
 * 
 * Posted on http://stackoverflow.com on 22 nov 2014
 * by http://stackoverflow.com/users/307099/stackoverflowuser2010
 * 
 *****************************************************************************/
public class EM {
    // Current estimated parameters.
    private Parameters _parameters;

    // Observations from the trials. These observations are set once.
    private final List<Observation> _observations;

    // Estimated observations per coin. These observations are the output
    // of the expectation step.
    private List<Observation> _expectedObservationsForCoinA;
    private List<Observation> _expectedObservationsForCoinB;

    private static java.io.PrintStream o = System.out;

    /*************************************************************************
     * Principal constructor.
     * 
     * @param observations
     *            The observations from the trial.
     * @param parameters
     *            The initial guessed parameters.
     *************************************************************************/
    public EM(List<Observation> observations, Parameters parameters) {
        _observations = observations;
        _parameters = parameters;
    }

    /*************************************************************************
     * Run EM until parameters converge.
     *************************************************************************/
    public Parameters run() {

        while (true) {
            expectation();

            Parameters estimatedParameters = maximization();

            o.printf("%s\n", estimatedParameters);

            if (_parameters.converged(estimatedParameters)) {
                break;
            }

            _parameters = estimatedParameters;
        }

        return _parameters;

    }

    /*************************************************************************
     * Given the observations and current estimated parameters, compute new
     * estimated completions (distribution over the classes) and observations.
     *************************************************************************/
    private void expectation() {

        _expectedObservationsForCoinA = new ArrayList<Observation>();
        _expectedObservationsForCoinB = new ArrayList<Observation>();

        for (Observation observation : _observations) {
            int numHeads = (int) observation.getNumHeads();
            int numTails = (int) observation.getNumTails();

            double probabilityOfObservationForCoinA =
                    binomialProbability(10, numHeads, _parameters.getThetaA());

            double probabilityOfObservationForCoinB =
                    binomialProbability(10, numHeads, _parameters.getThetaB());

            double normalizer =
                    probabilityOfObservationForCoinA
                            + probabilityOfObservationForCoinB;

            // Compute the completions for coin A and B (i.e. the probability
            // distribution of the two classes, summed to 1.0).

            double completionCoinA =
                    probabilityOfObservationForCoinA / normalizer;
            double completionCoinB =
                    probabilityOfObservationForCoinB / normalizer;

            // Compute new expected observations for the two coins.

            Observation expectedObservationForCoinA =
                    new Observation(numHeads * completionCoinA, numTails
                            * completionCoinA);

            Observation expectedObservationForCoinB =
                    new Observation(numHeads * completionCoinB, numTails
                            * completionCoinB);

            _expectedObservationsForCoinA.add(expectedObservationForCoinA);
            _expectedObservationsForCoinB.add(expectedObservationForCoinB);
        }
    }

    /*************************************************************************
     * Given new estimated observations, compute new estimated parameters.
     *************************************************************************/
    private Parameters maximization() {

        double sumCoinAHeads = 0.0;
        double sumCoinATails = 0.0;
        double sumCoinBHeads = 0.0;
        double sumCoinBTails = 0.0;

        for (Observation observation : _expectedObservationsForCoinA) {
            sumCoinAHeads += observation.getNumHeads();
            sumCoinATails += observation.getNumTails();
        }

        for (Observation observation : _expectedObservationsForCoinB) {
            sumCoinBHeads += observation.getNumHeads();
            sumCoinBTails += observation.getNumTails();
        }

        return new Parameters(sumCoinAHeads / (sumCoinAHeads + sumCoinATails),
                sumCoinBHeads / (sumCoinBHeads + sumCoinBTails));

        // o.printf("parameters: %s\n", _parameters);

    }

    /*************************************************************************
     * Since the coin-toss experiment posed in this article is a Bernoulli
     * trial, use a binomial probability Pr(X=k; n,p) = (n choose k) * p^k *
     * (1-p)^(n-k).
     *************************************************************************/
    private static double binomialProbability(int n, int k, double p) {
        double q = 1.0 - p;
        return nChooseK(n, k) * Math.pow(p, k) * Math.pow(q, n - k);
    }

    private static long nChooseK(int n, int k) {
        long numerator = 1;

        for (int i = 0; i < k; i++) {
            numerator = numerator * n;
            n--;
        }

        long denominator = factorial(k);

        return (long) (numerator / denominator);
    }

    private static long factorial(int n) {
        long result = 1;
        for (; n > 0; n--) {
            result = result * n;
        }

        return result;
    }

    /*************************************************************************
     * Entry point into the program.
     *************************************************************************/
    public static void main(String argv[]) {
        // Create the observations and initial parameter guess
        // from the (Do and Batzoglou, 2008) article.

        List<Observation> observations = new ArrayList<Observation>();
        observations.add(new Observation("HTTTHHTHTH"));
        observations.add(new Observation("HHHHTHHHHH"));
        observations.add(new Observation("HTHHHHHTHH"));
        observations.add(new Observation("HTHTTTHHTT"));
        observations.add(new Observation("THHHTHHHTH"));

        Parameters initialParameters = new Parameters(0.6, 0.5);

        EM em = new EM(observations, initialParameters);

        Parameters finalParameters = em.run();

        o.printf("Final result:\n%s\n", finalParameters);
    }
}


/*****************************************************************************
 * This class encapsulates the parameters of the problem. For this problem posed
 * in the article by (Do and Batzoglou, 2008), the parameters are thetaA and
 * thetaB, the probability of a coin coming up heads for the two coins A and B.
 *****************************************************************************/
class Parameters {
    double _thetaA = 0.0; // Probability of heads for coin A.
    double _thetaB = 0.0; // Probability of heads for coin B.

    double _delta = 0.00001;

    public Parameters(double thetaA, double thetaB) {
        _thetaA = thetaA;
        _thetaB = thetaB;
    }

    /*************************************************************************
     * Returns true if this parameter is close enough to another parameter
     * (typically the estimated parameter coming from the maximization step).
     *************************************************************************/
    public boolean converged(Parameters other) {
        if (Math.abs(_thetaA - other._thetaA) < _delta
                && Math.abs(_thetaB - other._thetaB) < _delta) {
            return true;
        }

        return false;
    }

    public double getThetaA() {
        return _thetaA;
    }

    public double getThetaB() {
        return _thetaB;
    }

    @Override
    public String toString() {
        return String.format("thetaA = %.5f, thetaB = %.5f", _thetaA, _thetaB);
    }

}

/*****************************************************************************
 * This class encapsulates an observation, that is the number of heads and tails
 * in a trial. The observation can be either (1) one of the observed
 * observations, or (2) an estimated observation resulting from the expectation
 * step.
 *****************************************************************************/
class Observation {
    double _numHeads = 0;
    double _numTails = 0;

    public Observation(String s) {
        for (int i = 0; i < s.length(); i++) {
            char c = s.charAt(i);

            if (c == 'H') {
                _numHeads++;
            } else if (c == 'T') {
                _numTails++;
            } else {
                throw new RuntimeException("Unknown character: " + c);
            }
        }
    }

    public Observation(double numHeads, double numTails) {
        _numHeads = numHeads;
        _numTails = numTails;
    }

    public double getNumHeads() {
        return _numHeads;
    }

    public double getNumTails() {
        return _numTails;
    }

    @Override
    public String toString() {
        return String.format("heads: %.1f, tails: %.1f", _numHeads, _numTails);
    }

}
